//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdlib.h>
#include <dstring.h>
#include "main.h"
#include "child.h"
#pragma hdrstop
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TChild *Child;
//---------------------------------------------------------------------------

void __fastcall TChild::FormCreate(TObject *Sender)
{
	// empty all bands
	for(int i = 0; i < DataGrid->ColCount; i++)
	{
		StringGrid1->Cells[i][0] = " ";
		StringGrid2->Cells[i][0] = " ";
		DataGrid->Cells[i][0] = " ";
	}
}
//---------------------------------------------------------------------------

void __fastcall TChild::FormClose(TObject *Sender, TCloseAction &Action)
{
    Action = caFree;	// fix silly windoze design flaw
			// yes, I called Close() and actually mean Close()... sigh.
}
//---------------------------------------------------------------------------

void __fastcall TChild::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	CanClose = true;

	// doesn't need to be saved
	if(!modified)
		return;

	switch(Application->MessageBox(("Do you want to save the changes you made to \""+Caption+"\"?").c_str(), "Turing Simulator", MB_YESNOCANCEL|MB_ICONEXCLAMATION))
	{
	case IDYES:
		Main->SaveActionExecute(Sender);
		break;
	case IDCANCEL:
		CanClose = false;
	// IDNO - nop
	}
}
//---------------------------------------------------------------------------

// Show the current state information from StateListView
// in each fields' respective control
void __fastcall TChild::StateListViewSelectItem(TObject *Sender,
      TListItem *Item, bool Selected)
{
	if(!Selected)
		return;

	current_item = Item;
	StateComboBox->Text = Item->Caption;
	InputEdit->Text = Item->SubItems->Strings[0];
	OutputEdit->Text = Item->SubItems->Strings[1];

	switch(Item->SubItems->Strings[2][1])
	{
	case 'L':
		LeftSpeedButton->Down = true;
		break;
	case 'R':
		RightSpeedButton->Down = true;
		break;
	case 'S':
		StopSpeedButton->Down = true;
		break;
	default:
		// all buttons off
		LeftSpeedButton->Down = true; // HACK: turn stop/right buttons off
		LeftSpeedButton->Down = false;
	}

	NewStateComboBox->Text = Item->SubItems->Strings[3];
}
//---------------------------------------------------------------------------

// One of the state input controls changed;
// update StateListView to reflect the changes
void __fastcall TChild::StateChange(TObject *Sender)
{
	if(!current_item)
		return;

	modified = true;

	switch(dynamic_cast<TComponent*>(Sender)->Tag)
	{
	// State changed
	case 0:
		current_item->Caption = StateComboBox->Text;
		break;

	// Input changed
	case 1:
		current_item->SubItems->Strings[0] = InputEdit->Text;
		break;

	// Output changed
	case 2:
		current_item->SubItems->Strings[1] = OutputEdit->Text;
		break;

	// New state changed
	case 3:
		current_item->SubItems->Strings[3] = NewStateComboBox->Text;
		break;

	// Movement changed
	// If the sender is Down, then Tag contains the ASCII code of the button;
	// otherwise, no button is pressed; return "N"
	default:
		TSpeedButton* button = dynamic_cast<TSpeedButton*>(Sender);
		if(button->Down)
			current_item->SubItems->Strings[2] = String((char)button->Tag);
		else
			current_item->SubItems->Strings[2] = "N";

		NewStateComboBox->SetFocus();
	}
}
//---------------------------------------------------------------------------

void __fastcall TChild::NewButtonClick(TObject *Sender)
{
	modified = true;

	current_item = StateListView->Items->Add();
	current_item->Caption = "0";
	current_item->SubItems->CommaText = "\"###\", \"###\", \"N\", \"0\"";
	StateListViewSelectItem(0, current_item, true);
}
//---------------------------------------------------------------------------

void __fastcall TChild::CopyButtonClick(TObject *Sender)
{
	if(!current_item)
		return;

	modified = true;

	TListItem* item = StateListView->Items->Add();
	item->Caption = current_item->Caption;
	item->SubItems->Assign(current_item->SubItems);
}
//---------------------------------------------------------------------------

void __fastcall TChild::DelButtonClick(TObject *Sender)
{
	if(!StateListView->Selected || !current_item)
		return;

	modified = true;

	StateListView->Selected->Delete();
	current_item = NULL;
}
//---------------------------------------------------------------------------

// Make sure the user enters only numbers in the state combo boxes
void __fastcall TChild::StateComboBoxKeyPress(TObject *Sender, char &Key)
{
	if(isdigit(Key) || Key == VK_RETURN || Key == VK_BACK);
	else
	{
		Beep();
		Key = 0;
	}
}
//---------------------------------------------------------------------------

// Add the state the user just entered to both combo box lists
void __fastcall TChild::StateComboBoxExit(TObject *Sender)
{
	TComboBox* box = dynamic_cast<TComboBox*>(Sender);

	// Is state already in the combo list?
	for(int i = 0; i < box->Items->Count; i++)
		if(box->Items->Strings[i] == box->Text)
			return;		// KLUDGE

	// no - add it
	StateComboBox->Items->Add(box->Text);
	NewStateComboBox->Items->Add(box->Text);
}
//---------------------------------------------------------------------------

// Process a turing machine state.
// Called from RunButtonClick or SpeedButton2Click.
// Reads input from the bands, and searces for a matching state.
// If no matching state is found, 'pause' execution (allowing the user to edit the bands),
//   and show an error message.
// Otherwise, write output onto the bands, save ITER information to allow undo/redo,
//   and move the r/w head.
void __fastcall TChild::do_state()
{
	modified = true;

	// get what's on the bands
	input = String(StringGrid1->Cells[grid_pos][0]) + String(StringGrid2->Cells[grid_pos][0]) + String(DataGrid->Cells[grid_pos][0]);

	// look for a matching state
	for(int i = 0; i < StateListView->Items->Count; i++)
	{
		TListItem* item = StateListView->Items->Item[i];
		String state_input = item->SubItems->Strings[0];

		// fix input wildcards (KLUDGE)
		for(int j = 1; j <= 3; j++)	// *&$%! 1-based string index
			if(state_input[j] == '#')
				state_input[j] = input[j];

		// got the right state? if not, check the next state
		if(StrToInt(item->Caption) != state || state_input != input)
			continue;

		// found the correct state

		// add new ITER to the ITER stack
		ITER* iter = (ITER*)malloc(sizeof(ITER));
		iter->next = iter_list;
		iter->prev_state = state;
		iter_list = iter;
		iter->input = new char[input.Length()+1];
		strcpy(iter->input, input.c_str());

		// get & process output
		String state_output = item->SubItems->Strings[1];

		// fix output wildcards (KLUDGE)
		for(int j = 1; j <= 3; j++)
			if(state_output[j] == '#')
				state_output[j] = input[j];

		// write output
		StringGrid1->Cells[grid_pos][0] = String(state_output[1]);
		StringGrid2->Cells[grid_pos][0] = String(state_output[2]);
		DataGrid->Cells[grid_pos][0] = String(state_output[3]);

		// may be changed to -1 if movement == 'S' so pre-set it
		state = StrToInt(item->SubItems->Strings[3]);

		int move = 0; // Default: N => don't move r/w heads
		String temp = item->SubItems->Strings[2];
		if(temp == "L")		// move left
			move = -1;
		else if(temp == "R")	// move right
			move = 1;
		else if(temp == "S")
			state =  -1;	// hard stop, can't run/single step past here

		// adjust grid_pos, expanding grid sizes, if necessary
		grid_pos += move;
		// grid_pos is out of bounds; increase grid size & init with " "
		if(grid_pos >= DataGrid->ColCount)
		{
			for(int i = DataGrid->ColCount; i < DataGrid->ColCount + 100; i++)
			{
				StringGrid1->Cells[i][0] = " ";
				StringGrid2->Cells[i][0] = " ";
				DataGrid->Cells[i][0] = " ";
			}

			StringGrid1->ColCount += 100;
			StringGrid2->ColCount += 100;
			DataGrid->ColCount += 100;
		}

		// set band position (show where the read/write head is)
		StringGrid1->Col = grid_pos;
		StringGrid2->Col = grid_pos;
		DataGrid->Col = grid_pos;

		iter->move = move;
		return;
	}

	Main->show_error("Failed to find matching state for state=[%d], input=[%s], grid_pos=[%d]!\nCheck the states list or edit the band contents before continuing", state, input.c_str(), grid_pos);
	RunButton->Tag = STOP; // stop at the beginning of the next state iteration
}
//---------------------------------------------------------------------------

// SINGLE STEP BACK
void __fastcall TChild::SpeedButton1Click(TObject *Sender)
{
	// don't undo beyond start
	ITER* iter = iter_list;
	if(!iter)
	{
		Beep();
		return;
	}

	modified = true;

	// restore band pos & data
	grid_pos -= iter->move;
	StringGrid1->Cells[grid_pos][0] = String(iter->input)[1];
	StringGrid2->Cells[grid_pos][0] = String(iter->input)[2];
	DataGrid->Cells[grid_pos][0] = String(iter->input)[3];

	// show band pos
	StringGrid1->Col = grid_pos;
	StringGrid2->Col = grid_pos;
	DataGrid->Col = grid_pos;

	state = iter->prev_state;

	// remove iter from list
	iter_list = iter->next;
	delete[] iter->input;
	free(iter);
}
//---------------------------------------------------------------------------

// SINGLE STEP FORWARD
void __fastcall TChild::SpeedButton2Click(TObject *Sender)
{
	// don't single step past end
	if(state == -1)
		Beep();
	else
		do_state();
}
//---------------------------------------------------------------------------

// Run the turing machine until a hard stop (state = -1), or RunButton->Tag == STOP
void __fastcall TChild::RunButtonClick(TObject *Sender)
{
	// user clicked 'RUN'
	if(RunButton->Caption == "&Run")
	{
		RunButton->Kind = bkNo;
		RunButton->Caption = "&Stop";
		Animate->Active = true;
		Main->SaveAction->Enabled = false;

		// run
		RunButton->Tag = RUNNING;
		while(state != -1 && RunButton->Tag == RUNNING)
		{
			do_state();
			Application->ProcessMessages(); // allow app update
		}

		Main->SaveAction->Enabled = true;
		Animate->Active = false;
		RunButton->Kind = bkYes;
		RunButton->Caption = "&Run";
	}
	// user clicked 'STOP'
	else
		RunButton->Tag = STOP;
}
//---------------------------------------------------------------------------

// move r/w heads to 0, and clear ITER list
void __fastcall TChild::ResetButtonClick(TObject *Sender)
{
	modified = true;

	state = 0;

	grid_pos = 0;
	StringGrid1->Col = 0;
	StringGrid2->Col = 0;
	DataGrid->Col = 0;	// can't do a=b=c=0 w/ properties, grr

	// free ITER list
	ITER* iter;
	while((iter = iter_list) != NULL)
	{
		iter_list = iter->next;
		delete[] iter->input;
		free(iter);
	}
}
//---------------------------------------------------------------------------


void __fastcall TChild::GridKeyPress(TObject *Sender, char &Key)
{
	modified = true;
	TStringGrid* grid = dynamic_cast<TStringGrid*>(Sender);
	grid->Cells[grid->Row,grid->Col][0] = Key; 
	grid->Col++; // next column has input focus, allows easy band filling
}
//---------------------------------------------------------------------------

// restrict input to 1 char
// (PARANOIA, only necessary for multi-char paste to cell)
void __fastcall TChild::GridGetEditMask(TObject *Sender, int ACol,
      int ARow, AnsiString &Value)
{
	Value = "c;1;_";
}
//---------------------------------------------------------------------------

static TStringGrid* clearme;

// Clear entire band
void __fastcall TChild::ClearBandClick(TObject *Sender)
{
	for(int i = 0; i < DataGrid->ColCount; i++)	// all bands are same size!
		clearme->Cells[i][0] = " ";		// HACK: last grid to have focus is cleared
}
//---------------------------------------------------------------------------

void __fastcall TChild::GridEnter(TObject *Sender)
{
	clearme = dynamic_cast<TStringGrid*>(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TChild::ClearListClick(TObject *Sender)
{
	StateListView->Items->Clear();
}
//---------------------------------------------------------------------------

