//---------------------------------------------------------------------------

#include <vcl.h>
#include "main.h"
#include "child.h"
#pragma hdrstop
#include <stdio.h>                     
//---------------------------------------------------------------------------

#pragma package(smart_init)
#pragma resource "*.dfm"
TMain *Main;

UINT __cdecl TMain::show_error(char* format, ...)
{
	char msg[0x100];

	va_list args;
	va_start(args, format);
	vsprintf(msg, format, args);
	va_end(args);

	return Application->MessageBox(msg, "Turing Simulator", MB_OK | MB_ICONEXCLAMATION);
}



void __fastcall TMain::ApplicationEventsHint(TObject *Sender)
{
	StatusBar->Panels->Items[0]->Text = (Application->Hint != "")? Application->Hint : (String)"Ready";
}
//---------------------------------------------------------------------------

void __fastcall TMain::ApplicationEventsIdle(TObject *Sender, bool &Done)
{
	if (MDIChildCount >= 1)
		child = dynamic_cast<TChild*>(ActiveMDIChild);
	else
		child = NULL;

	// TODO: idle enable/disable
	if(child)
	{
		SaveAction->Enabled = true;

		bool can_run = child->StateListView->Items->Count != 0;
		child->RunButton->Enabled = can_run;
		child->SpeedButton1->Enabled = can_run;
		child->SpeedButton2->Enabled = can_run;
	}
	else
	{
		SaveAction->Enabled = false;
	}
}
//---------------------------------------------------------------------------



void __fastcall TMain::NewActionExecute(TObject *Sender)
{
	child = new TChild(this);
	child->Caption = Format("Turing Machine%d", ARRAYOFCONST(((int)doc_counter)));
	doc_counter++;
}
//---------------------------------------------------------------------------

// open a tm, really belongs in child.cpp
void __fastcall TMain::OpenActionExecute(TObject *Sender)
{
	int i;

	if(!OpenDialog->Execute())
		return;

	child = new TChild(this);
	FILE* file;

	// load bands
	try
	{
		child->StringGrid1->Rows[0]->LoadFromFile(OpenDialog->FileName + ".1");
		child->StringGrid2->Rows[0]->LoadFromFile(OpenDialog->FileName + ".2");
		child->DataGrid->Rows[0]->LoadFromFile(OpenDialog->FileName + ".data");

		file = fopen(OpenDialog->FileName.c_str(), "r");
	}
	catch(...)
	{
		show_error("File (or an associated band file) not found or I/O error! Open aborted.");
		child->Close();
		return;
	}

	// If the first thing the user does after opening the program is to
	// open a file, close the 1st child window automatically created
	TChild* tmp = dynamic_cast<TChild*>(ActiveMDIChild);
	if(tmp->Caption == "Turing Machine1" && !tmp->modified)
	{
		tmp->Close();
		doc_counter--;
	}

	child->Caption = ExtractFileName(OpenDialog->FileName);
	child->named = true;
	SaveDialog->FileName = OpenDialog->FileName;

	char buf[0x100];

	// load turing machine states
	fscanf(file, "%d\n", &i);
	while(i--)
	{
		fscanf(file, "%s\n", &buf);
		TListItem* Item = child->StateListView->Items->Add();
		Item->Caption = String(buf);

		// can't use fscanf b/c buf will probably contain \0 char(s)
		fgets(buf, 0x100, file);
		buf[strlen(buf)-1] = '\0'; // strip '\n' that fgets reads in
		Item->SubItems->CommaText = String(buf);
	}

	// load simulator state
	fscanf(file, "%d %d\n", &(child->grid_pos), &(child->state));

	// load ITERs
	ITER* temp;
	ITER** head = &child->iter_list;

	fscanf(file, "%d\n", &i);
	while(i--)
	{
		temp = (ITER*)malloc(sizeof(ITER));
		temp->next = NULL;
		temp->input = new char[5];
		fscanf(file, "%d/%d/", &temp->prev_state, &temp->move);
		// can't use fscanf b/c temp->input will probably contain \0 char(s)
		fgets(temp->input, 0x100, file);
		temp->input[3] = '\0'; // strip '\n' that fgets reads in

		*head = temp;
		head = &(temp->next);
	}

	fclose(file);

	// use simulator state info to restore the 'operating environment'

	// show r/w head pos
	child->StringGrid1->Col = child->grid_pos;
	child->StringGrid2->Col = child->grid_pos;
	child->DataGrid->Col = child->grid_pos;
}
//---------------------------------------------------------------------------

// save a tm - should be in child.cpp
void __fastcall TMain::SaveActionExecute(TObject *Sender)
{
	int i;

	if(!child->named)
	{
		FileSaveAsClick(Sender);
		return;
	}

	child->Caption = ExtractFileName(SaveDialog->FileName);

	int num_items = child->StateListView->Items->Count;

	child->StringGrid1->Rows[0]->SaveToFile(SaveDialog->FileName + ".1");
	child->StringGrid2->Rows[0]->SaveToFile(SaveDialog->FileName + ".2");
	child->DataGrid->Rows[0]->SaveToFile(SaveDialog->FileName + ".data");

	FILE* file = fopen(SaveDialog->FileName.c_str(), "w");

	// save turing machine states
	fprintf(file, "%d\n", num_items);
	for(i = 0; i < num_items; i++)
	{
		fprintf(file, "%s\n", child->StateListView->Items->Item[i]->Caption.c_str());
		fprintf(file, "%s\n", child->StateListView->Items->Item[i]->SubItems->CommaText.c_str());
	}

	// save simulator state
	fprintf(file, "%d %d\n", child->grid_pos, child->state);

	// get number of ITERs to save
	ITER* temp = child->iter_list;
	for(i = 0; temp; temp = temp->next, i++);
	fprintf(file, "%d\n", i);

	// save ITERs
	temp = child->iter_list;
	while(i--)
	{
		fprintf(file, "%d/%d/%s\n", temp->prev_state, temp->move, temp->input);
		temp = temp->next;
	}

	fclose(file);

	child->modified = false;
}
//---------------------------------------------------------------------------

void __fastcall TMain::PrintActionExecute(TObject *Sender)
{
	// TODO:
}
//---------------------------------------------------------------------------

void __fastcall TMain::AboutActionExecute(TObject *Sender)
{
	Application->MessageBox("Copyright  2001 Jan Wassenberg", "About Turing Simulator", MB_OK|MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------



void __fastcall TMain::FileMenuClick(TObject *Sender)
{
	FileSaveAs->Enabled = child;
	FileClose->Enabled = child;
}
//---------------------------------------------------------------------------

void __fastcall TMain::FileCloseClick(TObject *Sender)
{
	child->Close();
}
//---------------------------------------------------------------------------

void __fastcall TMain::FileSaveAsClick(TObject *Sender)
{
	if(SaveDialog->Execute())
	{
		child->named = true;
		SaveActionExecute(0);
	}
}
//---------------------------------------------------------------------------

void __fastcall TMain::FileExitClick(TObject *Sender)
{
	Close();
}
//---------------------------------------------------------------------------

void __fastcall TMain::ViewToolbarClick(TObject *Sender)
{
	ViewToolbar->Checked = !ViewToolbar->Checked;
	CoolBar->Visible = ViewToolbar->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TMain::ViewStatusBarClick(TObject *Sender)
{
	ViewStatusBar->Checked = !ViewStatusBar->Checked;
	StatusBar->Visible = ViewStatusBar->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TMain::WindowMenuClick(TObject *Sender)
{
	CascadeAction->Enabled = child;
	TileAction->Enabled = child;
	ArrangeAction->Enabled = child;
}
//---------------------------------------------------------------------------

void __fastcall TMain::ToolbarPopupMenuPopup(TObject *Sender)
{
	PopUpToolbar->Checked = ViewToolbar->Checked;
	PopUpStatusBar->Checked = ViewStatusBar->Checked;
}
//---------------------------------------------------------------------------

